# Created by honza@tnx.cz
# Created during lot of winter nights and weekend afternoons
# It is far from professional work, but it does the job. May need some optimization for larger data.
# Feel free to use, change, distribute

# We are going to need some geolocation servis. This script uses https://ipapi.com/
# You can generate your API key for free. It is enough for most usage cases, but there is a limit for requests per day. So several people using the same key could go over quota.
# As I do not want to reveal my API KEY accidentaly while sharing the script, I have saved it in the text file and I am reading it from it.
# Do not forget to put ? in front of your API KEY if you want to use it the same way as I do.
$apikey = Get-Content .\apikey.txt

# Every day at 6:00 AM we are going to read current IIS Log and log from previous day. Attackers are nocturnal animals, lets have some overlap of logs.
# It may be used for any service using IIS after some changes. I have only Exchange server runing.

# For testing purposes when we use older logs, we may want to set the variables manually.
# $today = "220315"
# $yesterday = "220314"
# You may also want to edit your IIS log and put ther an IP address that is located far away, so that you can see if it works for you.

$today = get-date -Format yyMMdd
$yesterday = ((get-date).AddDays(-1)).tostring('yyMMdd')

# Now put the file names together
$logtoday = ".\u_ex" + $today + ".log"
$logyesterday = ".\u_ex" + $yesterday + ".log"
$results_path = ".\" + $today + "_results.txt"
$distance_path = ".\" + $today + "_max_distance.txt"
$speed_path = ".\" + $today + "_max_speed.txt"
$check_path = ".\" + $today + "_max_check.txt"
"Username;Max_speed;Max_distance" > $check_path

# Get-Content gets the file, pipe to Where-Object and skip the first 3 lines. Because they look like this:
#Software: Microsoft Internet Information Services 10.0
#Version: 1.0
#Date: 2020-12-11 00:00:00

$Log1 = Get-Content $logyesterday | Where-Object {$_ -notLike "#[D,S-V]*" }
$Log2 = Get-Content $logtoday | Where-Object {$_ -notLike "#[D,S-V]*" }

# Replace unwanted text in the line containing the columns. We are reading names of columns. We just need to read one of the files.
$Columns = (($Log1[0].TrimEnd()) -replace "#Fields: ", "" -replace "-","" -replace "\(","" -replace "\)","").Split(" ")

# Count available Columns, used later
$Count = $Columns.Length

# Get all Rows that I Want to retrieve, only successful connections and specific services
$Rows1 = $Log1 | Where-Object {($_ -like "*200*") -and (($_ -like "*Microsoft-Server-ActiveSync*") -or ($_ -like "*EWS*") -or ($_ -like "*OWA*"))}
$Rows2 = $Log2 | Where-Object {($_ -like "*200*") -and (($_ -like "*Microsoft-Server-ActiveSync*") -or ($_ -like "*EWS*") -or ($_ -like "*OWA*"))}
$Rows = $Rows1 + $Rows2

# Create an instance of a System.Data.DataTable
$IISLog = New-Object System.Data.DataTable "IISLog"

# Loop through each Column, create a new column through Data.DataColumn and add it to the DataTable. Everything is a string. This will be important later.
foreach ($Column in $Columns) {
 $NewColumn = New-Object System.Data.DataColumn $Column, ([string])
 $IISLog.Columns.Add($NewColumn)
}

# Loop through each Row and add the Rows.
foreach ($Row in $Rows) {
 $Row = $Row.Split(" ")
 $AddRow = $IISLog.newrow()
for($i=0;$i -lt $Count; $i++) {
 $ColumnName = $Columns[$i]
 $AddRow.$ColumnName = $Row[$i]
 }
 $IISLog.Rows.Add($AddRow)
}
# $IISLog is ready to be used.

# Creating another data table for results
$res_cols = @("username","IP1","Country1","IP2","Country2","Distance","Time","Speed","Severity")
$results = New-Object System.Data.DataTable "Results"
foreach ($Col in $res_cols) {
 $NewColumn = New-Object System.Data.DataColumn $Col, ([string])
 $results.Columns.Add($NewColumn)
}

# Creating another two data tables for shorter results based on speed and distance
$speed_cols = @("username","IP1","Country1","IP2","Country2","Distance","Time","Speed")
$speed_results = New-Object System.Data.DataTable "Speed_Results"
foreach ($Col in $speed_cols) {
 $NewColumn = New-Object System.Data.DataColumn $Col, ([string])
 $speed_results.Columns.Add($NewColumn)
}

$distance_cols = @("username","IP1","Country1","IP2","Country2","Distance","Time","Speed")
$distance_results = New-Object System.Data.DataTable "Distance_Results"
foreach ($Col in $distance_cols) {
 $NewColumn = New-Object System.Data.DataColumn $Col, ([string])
 $distance_results.Columns.Add($NewColumn)
}

# Creating shorter version of log without localhost entries, we need only four columns.
$shortlog = $iislog | Where-Object {($_.csusername -ne '-') -and ($_.cip -ne '127.0.0.1') -and ($_.cip -ne '::1')} | Select-Object date, time, csusername, cip

# We need to fix usernames so that we have same format and we need to fix private addresses
foreach ($row in $shortlog) {
 $ip = $row.cip
 $user = $row.csusername
 $user = $user.tolower()
# Private addresses cannot get a proper location, so we change them to our external public IP address 
  if (($ip -like '192.*') -or ($ip -like '10.*') -or ($ip -like '172.*')) {
  $ip = '188.244.55.168'
# Return new address to the object.
  $row.cip = $ip
 }
# Users are able to do horrible stuff with credentials. Someone uses netbios domain name, someone uses UPN. Lets keep samaccountname in lower case and strip anything else.
# These are my domains, you need to replace the strings for your domains
$user = $user -replace "@tnx.cz","" -replace "tnx.cz\\","" -replace "tnx\\","" -replace "@jan-kovar.cz",""
# Return username to the object.
 $row.csusername = $user
}

# With the free IPDB account we have limited ammount of querries. We can easily deplete our quota for testing. So we are going to store the results for each IP address localy.
# We are going to search locally first and if it is not in our DB, we will contact online DB.
# Reading database of IP address locations
#
$ipfile = import-csv .\ip_database.txt -delimiter ','
# Creating hash table of IP addresses, IP address is primary key. Reading the CSV into the Hash Table.
$ipdb = @{}
foreach ($line in $ipfile) {
 $ipdb[$line.key]=$line.value
}

# Creating list of distinct IP addresses from log. IP address can be listed multiple times in the log, but the location is always the same so we need to check just once.
$iplist = $shortlog |  Select-Object cip | sort-object -Property cip | Get-Unique -AsString

# We check the IP against our database. If it is present, we do not need to do anything. If it is not present, we need to check online database and add results to our database.
foreach ($address in $iplist) {
 $ip = $address.cip
 if ($ipdb.ContainsKey($ip)) {
  }
 else {
     $uri = "http://api.ipapi.com/" + $ip + $apikey
     $request = Invoke-RestMethod -Method Get -Uri $uri
     $Lat = $request.Latitude
     $Long = $request.Longitude
     $co = $request.country_name
# We need latitude, longitude and country, put it together separated by semicolon and save it as value for the IP address key.     
     $value = "$lat" + ";" + "$long" + ";" + "$co"
     $ipdb.add($ip, $value)
  }
}

# Creating distinct list of usernames from log.
$users = $shortlog | Select-Object csusername | sort-object -Property csusername | Get-Unique -AsString

foreach ($user in $users) {
# Creating empty array list. Also deleting content in every loop.
         $collindex = [System.Collections.ArrayList]@()
         $name = $user.csusername

# Getting records for selected user
         $userrecords = $shortlog | Where-Object {$_.csusername -eq $name} | Sort-Object date, time

# We need helping variable $stillthesame. This part make processing a lot faster. We do not need to compare all records. If the user connected from the same IP address several 
# times in row, we just need the first and last record. Everything in between can be removed. But that is valid only for a group of connections that were not interupted by a 
# connection from other address. (AAAABBBBAAAACCCCAAAAADDDDD) makes (AABBAACCAADD)
         $stillthesame = 0
         foreach ($userrecord in $userrecords) {
              $ip1 = $userrecord.cip
# We need to know the position of the record in the array so that we can delete it if needed.          
              $index1 = $userrecords.indexof($userrecord)
# We need to know IP adress in next record, so that we can compare it. For that we need index + 1
              $index2 = $index1 + 1
              $ip2 = $userrecords[$index2].cip
# Compare IP addresses          
              if ($ip1 -eq $ip2) {
# If they are the same, but the variable $stillthesame is 0, it is first address in the line and we need to keep it. But we know that we have two same addresses and we change the variable to 1.
                   if ($stillthesame -eq 0) {
                        $stillthesame = 1
                   }
# If $stillthesame is already 1, then it means that in the previous record and next record is the same IP address, so this record can be deleted. We just add the index to the aray list.
# We are going to remove the records later.
                   else {
                        $collindex.add($index1) > $null
                        }
              }
# If the addresses were not the same, lets reset the help variable.
              else {
                   $stillthesame = 0
              }
              }
# We have gathered indexes of records that can be deleted because they contain redundand addresses. We need to delete them from the end. If we delete them from the start, position of next records
# changes and we would delete incorrect records. So we need to sort them in descending order.          
         $collindex = $collindex | Sort-Object -Descending
# We need to remove records from array and that is not usually possible because it has fixed size. We need to convert it to System.Collections.ObjectModel.Collection.     
         $final = {$userrecords}.Invoke()
# Now we can remove the records on indexes that we gathered.     
         foreach ($index in $collindex) {
              $final.removeat($index)
         }
         
         $collection1 = $final
# Making system table from array - will be needed later to remove first row.
         $collection2 = $collection1
# We will need to remove records from collection, so again we are converting it to System.Collections.ObjectModel.Collection.
         $collection2 = {$collection2}.invoke()
     
# Comparing each entry from collection1 with all following entries in the collection
# We made a copy of collection. They are the same. So it does not make sense to compare same records from the collections.  We also do not want to repeat the comparison.
# Example: Collection has values A, B, C. We have created a copy and want to compare: A-B, A-C, B-C. So after loop of collection 1, we need to remove first line in collection 2
# When we get to the last record in collection 1, collection 2 would be empty. We need to cover that.
      foreach ($collrow1 in $collection1) {
       $ip1 = $collrow1.cip
       $date1 = $collrow1.date
       $time1 = $collrow1.time
# Creating date + time variable
       $dt1 = $date1 + " " + $time1
# For the last record in collection 1, collection 2 would be empty. So lets check, that it is not.
       if ($null -ne $collection2) {
# Removing first row from second collection. They are the same. We want to start comparing first two rows not the same row. First row will be removed in each cycle. We want to compare: A-B. A-C, B-C.
        $collection2.RemoveAt(0)
# Compare all remaining records from collection 2 with selected record from collection 1
        foreach ($collrow2 in $collection2) {
         $ip2 = $collrow2.cip
         $date2 = $collrow2.date
         $time2 = $collrow2.time
# Creating date + time variable
         $dt2 = $date2 + " " + $time2
     
# Calculating time difference in hours.
         $howlong = (New-TimeSpan -Start $dt1 -End $dt2).totalhours
# Same IP addresses = same location, no reasong to compare. If they are different, lets calculate. We do not compare events more than 12 hours appart.
         if (($ip1 -ne $ip2) -and ($howlong -lt 12)) {
          $process1 = $ipdb[$ip1]
# $ipdb is a hash table, where key is IP address and value is latitude, longitude and country separated by semicolon      
          $processed1 = $process1 -split ";"
          $latitude1 = $processed1[0]
          $longitude1 = $processed1[1]
          $country1 = $processed1[2]
     
          $process2 = $ipdb[$ip2]
          $processed2 = $process2 -split ";"
          $latitude2 = $processed2[0]
          $longitude2 = $processed2[1]
          $country2 = $processed2[2]
# Math part begins. We have latitude and longitude data in degrees. The formula for calculating distance requires it in radians.
# Recalculate latitude and longitude in radians instead of degrees
          $latitude1 = ([math]::PI /180) * $latitude1
          $latitude2 = ([math]::PI /180) * $latitude2
          $longitude1 = ([math]::PI /180) * $longitude1
          $longitude2 = ([math]::PI /180) * $longitude2
# Calculating distance between coordinates. No I do not remember the formula. I had to find it.
          $distance = [math]::acos([math]::cos($latitude1)*[math]::cos($longitude1)*[math]::cos($latitude2)*[math]::cos($longitude2) + [math]::cos($latitude1)*[math]::sin($longitude1)*[math]::cos($latitude2)*[math]::sin($longitude2) + [math]::sin($latitude1)*[math]::sin($latitude2)) * 6371
# We know distance and time between connections. We can calculate speed how fast would user have to go to move from location 1 to location 2
          $speed = $distance / $howlong
# Anything under speed 100 km/hod and distance 300km can is not important. Increase severity with speed.
          $sev = "OK"
          if (($speed -gt 100) -and ($distance -gt 300)) {
           $sev="Low"
            if ($speed -gt 500) {
            $sev="Medium"
            if ($speed -gt 1000) {
             $sev="High"
            }
           }
# Mobile operators usually return location in capital city. If there is a big distance within country, severity is still low.                        
           if ($country1 -eq $country2) {
            $sev = "Low"
           }
# Some rounding so that we do not see to many decimal places. We keep 4 places for time, because it is in hours and if the log events were just second appart, we need 4 decimal places.
           $distance = [math]::Round($distance)
           $howlong = [math]::Round($howlong,4)
           $speed = [math]::Round($speed)
# If the results are concerning, we add the vaalues to our data table. If you want to upload it do SQL database, this should be the best object type.
            if ($sev -ne "OK") {
                $AddRow = $results.newrow()
                $AddRow.username = $name
                $AddRow.IP1 = $ip1
                $AddRow.Country1 = $country1
                $AddRow.IP2 = $ip2
                $AddRow.Country2 = $country2
                $AddRow.Distance = $distance
                $AddRow.Time = $howlong
                $AddRow.Speed = $speed
                $AddRow.Severity = $sev
                $results.Rows.Add($AddRow)
                }     
            }
         }
        }
       }
      }
    }


# We have finished processing our logs. There might have been some new IP addresses in the log, so we want to save them in our local IP database.
$ipdb.GetEnumerator() | Select-Object -Property key, value | Export-csv -NoTypeInformation -Delimiter ',' -Path .\ip_database.txt

# We are not done with processing of our findings, but we may want to save it to CSV.
$results | export-csv -NoTypeInformation $results_path

# We have lot of redundant data in the results, we want to clean them. If you connect once from China and ten times from home, you get 10 warnings. We just need one.
# Lets read what users have dangerous data in the logs
$users = $results | Select-Object username | sort-object -Property username | Get-Unique -AsString

# We are going to create two hash tables.
$distance_db = @{}
$speed_db = @{}

foreach ($user in $users) {
    $who = $user.username
# We just want results for current user
    $small_users = $results | Where-Object {$_.username -eq $who}
# This is only used for reference. We can check whether the maximum value is the same in other exported files.    
    $ms = ($small_users | Measure-Object -Property Speed -Maximum).Maximum
    $md = ($small_users | Measure-Object -Property Distance -Maximum).Maximum
    $text = $who + ";" + $ms + ";" + $md
    $text >> $check_path

    foreach ($line in $small_users) {
        $ip1 = $line.IP1
        $ip2 = $line.IP2
# The primary key for our hash table will be combination of username and IP addresses. We want to check whether we already checked this combination.        
        $hash_key1 = $who + ";" + $ip1 + ";" + $ip2
        $hash_key2 = $who + ";" + $ip2 + ";" + $ip1
# It does not matter if addresses A and B are as IP1 and IP2 or IP2 and IP1. The distance is the same. So we need to search for both keys. If it exist, we have already processed
# this combination of addresses and we do not need to do anything.
        if (($distance_db.ContainsKey($hash_key1)) -or ($distance_db.ContainsKey($hash_key2))) {

        }
        else {
# If we did not process this combination of addresses, we are going to filter all records with these two addresses, search for maximum Distance and sort it descending.
# It is actually stupid, because same comple of addresses will always have the same destination. But it will also sort out all redundant records. There will be just one.
# Also now is the time to remember that we have created the datatable object with all columns as string. But the distance is number, so we need to specify 
# that we want to sort it as Integer bz: Sort-Object {[int]$_.Distance}         
            $sort = @($small_users | Where-Object {(($_.IP1 -eq $ip1) -and ($_.IP2 -eq $ip2)) -or (($_.IP2 -eq $ip1) -and ($_.IP1 -eq $ip2))} | Sort-Object {[int]$_.Distance} -Descending)

            $ip_res1 = $sort[0].IP1
            $ip_res2 = $sort[0].IP2
            $co_res1 = $sort[0].Country1
            $co_res2 = $sort[0].Country2
            $dis_res = $sort[0].Distance
            $time_res = $sort[0].Time
            $speed_res = $sort[0].Speed
            $value = $who + ";" + $ip_res1 + ";" + $co_res1 + ";" + $ip_res2 + ";" + $co_res2 + ";" + $dis_res + ";" + $time_res + ";" + $speed_res
# And add it to the hash table so that we do not process it again.            
            $distance_db.add($hash_key1, $value)
            
# We do the same with Speed, but this time it makes sense. Distance is the same, but time differs so speed also differs. We are going to sort it descending, again we need to make sure
# that it is Integer.            
            $sort = @($small_users | Where-Object {(($_.IP1 -eq $ip1) -and ($_.IP2 -eq $ip2)) -or (($_.IP2 -eq $ip1) -and ($_.IP1 -eq $ip2))} | Sort-Object {[int]$_.Speed} -Descending)

            $ip_res1 = $sort[0].IP1
            $ip_res2 = $sort[0].IP2
            $co_res1 = $sort[0].Country1
            $co_res2 = $sort[0].Country2
            $dis_res = $sort[0].Distance
            $time_res = $sort[0].Time
            $speed_res = $sort[0].Speed
            $value = $who + ";" + $ip_res1 + ";" + $co_res1 + ";" + $ip_res2 + ";" + $co_res2 + ";" + $dis_res + ";" + $time_res + ";" + $speed_res
            $speed_db.add($hash_key1, $value)

        }
        
    }
}

# Both Hash Tables are difficult to read. They have complicated key and all results are in the same filed of value
# We are going to convert it to array and then split the value and load it into the datatable object.
# We export all files, so you can compare yymmdd_results.txt, yymmdd_max_distance.txt and yymmdd_max_speed.txt

# Convert Hash Table $distance_db to Array $distance_arr
$distance_arr = @($distance_db.GetEnumerator() | Select-Object -Property key, value)
foreach ($line in $distance_arr) {
# Split data into array
     $data = $line.value.Split(";")
# Add data to the datatable object     
     $AddRow = $distance_results.newrow()
     $AddRow.username = $data[0]
     $AddRow.IP1 = $data[1]
     $AddRow.Country1 = $data[2]
     $AddRow.IP2 = $data[3]
     $AddRow.Country2 = $data[4]
     $AddRow.Distance = $data[5]
     $AddRow.Time = $data[6]
     $AddRow.Speed = $data[7]
     $distance_results.Rows.Add($AddRow)
 }

# Export to CSV
$distance_results | Export-Csv -NoTypeInformation -Delimiter "," $distance_path

$speed_arr = @($speed_db.GetEnumerator() | Select-Object -Property key, value)
foreach ($line in $speed_arr) {
     $data = $line.value.Split(";")
     $AddRow = $speed_results.newrow()
     $AddRow.username = $data[0]
     $AddRow.IP1 = $data[1]
     $AddRow.Country1 = $data[2]
     $AddRow.IP2 = $data[3]
     $AddRow.Country2 = $data[4]
     $AddRow.Distance = $data[5]
     $AddRow.Time = $data[6]
     $AddRow.Speed = $data[7]
     $speed_results.Rows.Add($AddRow)
 }

# Export to CSV
$speed_results | Export-Csv -NoTypeInformation -Delimiter "," $speed_path
